; Pulse Sensor actor

; Node addresses:
;	100 - forwarder node address
;	200 - actor directory service address
;	101 - sensor 1
;	102 - sensor 2
;	103 - sensor 3

; Message types:
;	11 - pulse
;	12 - pulse collection
;	13 - pulse pair
;	14 - mobile object
;	15 - vehicle
;
;	20 - actor activation
;	21 - actor request
;	22 - actor directory registration
;	23 - actor directory query

; Constants:
;	44 - distance between adjacent sensors
;	 5 - threshold for vehicle speed
;	25 - threshold for vehicle length

; Message formats:
;	pulse
;		(list msg_id sensor_id start_time duration)
;	pulsecollection
;		(list msg_id pulsemsg pulsemsg pulsemsg)
;	pulsepair
;		(list msg_id pulsemsg pulsemsg)
;	mobileobject
;		(list msg_id speed length direction pulsepairmsg)
;	vehicle
;		(list msg_id mobileobjectmsg)
;	actor activation
;		(list msg_id)
;	actor request
;		(list msg_id actorid)
;	actor directory registration
;		(list msg_id actorname actorid)

; top-level actor
(	(rec (vehicleidentifier reqlist arglist result getargs behavior sendresult)
	(seq

		; wait for activation
		(	(rec (waitforact)
				(cond (equal (cdr msgq) nil)
					(seq (sleep 100) (waitforact))
					(cond (not (equal (caadr msgq) 20))
						(seq (sleep 100) (waitforact))
						(print 20)	; received request, moving on
					)
				)
			)
		)

		; when activated, self-register
		(send 200 22 (list (quote vehicleidentifier) (id)))

		; act as a request receiving server
		(	(rec (waitforreq)
				(cond (equal (cdr msgq) nil)
					(seq (sleep 100) (waitforreq))
					(cond (not (equal (caadr msgq) 21))
						(seq (sleep 100) (waitforreq))
							(par 
								(cond (equal (cdr reqlist) nil)	; first req
									(seq
										(getargs nil nil)
										(behavior)
										(sendresult reqlist)
									)
									(list)
								)
								(seq
									(set reqlist (cons (caaddr msgq) reqlist))
									(sleep 100)
									(waitforreq)
								)
							)
						)
					)
				)
			)
		)
	)
	)

	nil	; reqlist

	nil	; arglist

	nil	; result

	; getargs - collect arguments needed for executing behavior
	(rec (getargs types ids)
		(seq
			; query ADS
			(	(lambda (queryads m) (queryads m))
				; wait for responses
				(rec (queryads m)
					(cond (equal m 0)
						()
						(cond (equal (cdr msgq) nil)
							(seq (sleep 100) (queryads m))
							(seq
								(set ids (cons (caddadr msgq) ids))
								(queryads m-1)
							)
						)
					)
				)
				; send ADS requests (and count their number
				(	(lambda (sendreq n) (sendreq n))
					(rec (sendreq n)
						(cond (equal types nil)
							n
							(seq
								(send 200 (list 23 (car l))) 
								(sendreq n+1)
							)
						)
					)
					0
				)
			)

			; send activation and request msgs to actors
			(	(lambda (sendreq l) (sendreq l))
				(rec (sendreq l)
					(cond (equal l nil)
						()
						(seq
							(send (car l) (list 20))
							(send (car l) (list 21 (id)))
							(sendreq (cdr l))
						)
					)
				)
				ids
			)

			; wait for replies
			(	(lambda (waitforargs l) (waitforargs l))
				(rec (waitforargs l)
					(cond (equal l nil)
						()
						(cond (equal (cdr msgq) nil)
							(seq (sleep 100) (waitforargs l))
							(seq
								(set arglist (cons (cadadr msgq) arglist))
								(waitforargs (cdr l))
							)
						)
					)
				)
				ids
			)
		)
	)

	; behavior - apply application logic to list of arguments, return result
	(rec (behavior)
		(set result (prim 5))
	)

	; sendresult - send the result to interested parties (reqlist)
	(rec (sendresult l)
		(cond (equal l nil)
			(print 333)	; done
			(seq
				(send (car l) result)
				(sendresult (cdr l))
			)
		)
	)
)
