/**********************************************************
	admclient.c
	version 1.1
	Kirill Mechitov (mechitov@cs.uiuc.edu)
**********************************************************/

/* Actor Deployment and Messaging Client */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <fcntl.h>
#include <pthread.h>

#define MAX_ACTOR_SIZE	10000
#define MAX_MSG_SIZE	10000

typedef struct {
	char **s;
	int i;
} thread_args;

static void errquit(char *msg)
{
	fprintf(stderr, "error: %s\n", msg);
	exit(1);
}

void *client_thread(void *args)
{
	thread_args *ta = (thread_args *)args;
	char **argv = ta->s;
#ifdef CONCUR
	int idx = ta->i;
#endif

	char *addr;
	int port;
	int sock;
	struct sockaddr_in sa;
	struct hostent *host;
	char actor[MAX_ACTOR_SIZE], actor_msg[MAX_MSG_SIZE];
#ifdef CONCUR
	char ma_id[20], code_name[20];
#else
	char *ma_id, *code_name;
#endif
	int int_ma_id, msg_size, alen;
	char ack, *eof = "-1";
	int f;

	srand(time(NULL));

	/* get admserver address and port number */
#ifdef CONCUR
	int_ma_id = atoi(argv[1]) + idx;
	sprintf(ma_id, "%d", int_ma_id);
	sprintf(code_name, "%s%d", argv[2], idx);
#else
	ma_id = argv[1];
	int_ma_id = atoi(argv[1]);
	code_name = argv[2];
#endif
	if ((f = open(code_name, O_RDONLY)) < 0) {
		errquit("open failed");
	}
	if ((alen = read(f, actor, MAX_ACTOR_SIZE)) < 0) {
		errquit("read failed");
	}
	actor[alen] = 0;
	close(f);

	/* connect to server */
	addr = argv[3];
	port = atoi(argv[4]);
	if ((host = gethostbyname(addr)) == NULL) {
		errquit("gethostbyname failed");
	}
	memset(&sa, 0, sizeof (sa));
	sa.sin_family = AF_INET;
	memcpy(&sa.sin_addr, host->h_addr, host->h_length);
	sa.sin_port = htons(port);
	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		errquit("socket failed");
	}
	if (connect(sock, (struct sockaddr*)&sa, sizeof (sa)) < 0) {
		errquit("connect failed");
	}

	/* send meta-actor id, wait for ack */
	if (write(sock, ma_id, strlen(ma_id)) < 0) {
		errquit("write failed");
	}
	if (read(sock, &ack, sizeof (char)) < 0 || ack != '1') {
		errquit("read ack failed");
	}
	read(sock, &ack, sizeof (char));

	/* send actor code */
	if (write(sock, actor, alen+1) < 0) {
		errquit("write failed");
	}
	actor[alen] = 0;
	actor[alen+1] = 0;
	if (write(sock, eof, 2) < 0) {
		errquit("write failed");
	}
	printf("%d: actor deployed\n", int_ma_id);

	/* wait for messages (blocking wait) */
	for (;;) {
		if ((msg_size = read(sock, actor_msg, MAX_MSG_SIZE)) < 0) {
			errquit("read failed");
		}
		actor_msg[msg_size] = 0;

		/* process message */
		actor_msg[msg_size] = 0;
		printf("%d: received actor msg %s\n", int_ma_id, actor_msg);

		/* stop execution and unregister */
		if (1) {
			break;
		}
	}

	/* close connection */
	if (write(sock, eof, 2) < 0) {
		errquit("write failed");
	}
	close(sock);

	pthread_exit(NULL);
	return NULL;
}

int main(int argc, char **argv)
{
	int i;
	thread_args ta[3];
	pthread_t thread[3];

	if(argc != 5) {
		fprintf(stderr, "usage: %s <ma_id> <actor> <address> <port>\n", argv[0]);
		exit(1);
	}

#ifdef CONCUR
	for (i = 0; i < 3; i++) {
#else
	for (i = 0; i < 1; i++) {
#endif
		ta[i].s = argv;
		ta[i].i = i;
		pthread_create(&thread[i], NULL, client_thread, &ta[i]);
	}

#ifdef CONCUR
	for (i = 0; i < 3; i++) {
#else
	for (i = 0; i < 1; i++) {
#endif
		pthread_join(thread[i], NULL);
	}

	return 0;
}
