/**********************************************************
	fwd.c
	version 1.1
	YoungMin Kwon (ykwon4@cs.uiuc.edu)
	Kirill Mechitov (mechitov@cs.uiuc.edu)
**********************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include "packet.h"

#define MaxClient	100

static void errquit(char *msg)
{
	fprintf(stderr, "error: %s\n", msg);
	exit(1);
}

void run(int port)
{
	int ssd;				/* server socket */
	int sd[MaxClient];		/* sockets */
	struct sockaddr_in ca, sa;
	fd_set rfd;

	memset(sd, 0xff, sizeof (int)*MaxClient);	/* initialized to -1 */

	memset(&sa, 0, sizeof (sa));
	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = htonl(INADDR_ANY);
	sa.sin_port = htons(port);

	if ((ssd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		errquit("socket failed");
	}
	if (bind(ssd, (struct sockaddr*) &sa, sizeof (sa)) < 0) {
		errquit("bind failed");
	}
	if(listen(ssd, 5) < 0) {
		errquit("listen failed");
	}

	for (;;) {
		int maxd = ssd;
		int	i, j, k, l;

		FD_ZERO(&rfd);
		FD_SET(ssd, &rfd);				/* for new connection */

		for (i = 0; i < MaxClient; i++) {
			if (sd[i] >= 0) {
				maxd = sd[i] > maxd ? sd[i] : maxd;
				FD_SET(sd[i], &rfd);	/* for existing connection */
			}
		}

		if (select(maxd + 1, &rfd, NULL, NULL, NULL) < 0) {
			errquit("select failed");
		}

		for (i = 0; i < MaxClient; i++) {
			if (sd[i] >= 0 && FD_ISSET(sd[i], &rfd)) {
				int n, tmp;
				char buf[PacketLength * 10];

				n = read(sd[i], buf, PacketLength * 10);

				fprintf(stderr, "%d:%d ", i, n);

				/* print packet */
				printf("\n");
				for (tmp = 0; tmp < n; tmp++) {
					printf("%d ", (unsigned char)buf[tmp]);
				}
				printf("\n");

				if (n <= 0) {			/* connection is closed */
					close(sd[i]);
					sd[i] = -1;
				}
				else {
					for (j = 0; j < MaxClient; j++) {
						if (j == i) {	/* prevent forwarding to the sender */
							continue;
						}
						if (sd[j] < 0) {
							continue;
						}
						for (k = 0; k < n; k += l) {
							if ((l = write(sd[j], buf + k, n - k)) < 0) {
								close(sd[j]);
								sd[j] = -1;
								break;
							}
						}
					}
				}
			}
		}

		if (FD_ISSET(ssd, &rfd)) {		/* new connection */
			fprintf(stderr, "! ");
			for (i = 0; i < MaxClient; i++) {
				if (sd[i] < 0) {
					sd[i] = accept(ssd, (struct sockaddr*) &ca, &i);
					break;
				}
			}
		}
	}
}

int main(int argc, char **argv)
{
	if (argc != 2) {
		fprintf(stderr, "usage: %s <port>\n", argv[0]);
		exit(0);
	}
	signal(SIGPIPE, SIG_IGN);
	run(atoi(argv[1]));	
	return 0;
}
