/**********************************************************
	parse.c
	version 1.1
	YoungMin Kwon (ykwon4@cs.uiuc.edu)
	Kirill Mechitov (mechitov@cs.uiuc.edu)
**********************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>

#include "memory.h"
#include "data.h"
#include "eval.h"
#include "comm.h"
#include "thread.h"
#include "hw.h"
#ifdef ADMSERVER
#include "admserver.h"
#endif

typedef struct SymTbl {
	char *sym;
	uint16_t adrs;
	struct SymTbl *next;
} SymTbl;
typedef struct NumTbl {
	int num;
	uint16_t adrs;
	struct NumTbl *next;
} NumTbl;
typedef struct {
	char c;
	int	fd;
	SymTbl st;
	NumTbl nt;
} Scan;
static uint16_t parse(int depth, Scan *sc);

static uint16_t lookupSym(SymTbl *p, char *s) {	//keywords...
	if(p->next==NULL) {
		p->sym=strdup(s);
		p->adrs=name(p->sym);
		p->next=(SymTbl*)calloc(1,sizeof(SymTbl));
		return p->adrs;
	}
	else if(!strcmp(p->sym, s))
		return p->adrs;
	else
		return lookupSym(p->next, s);
}
static uint16_t lookupNum(NumTbl *p, int n) {
	if(p->next==NULL) {
		p->num=n;
		p->adrs=newWord(n);
		p->next=(NumTbl*)calloc(1,sizeof(NumTbl));
		return p->adrs;
	}
	else if(p->num==n)
		return p->adrs;
	else
		return lookupNum(p->next, n);
}
static char get(Scan *sc) {
	char c;
	if(sc->c==0)  {
		if(read(sc->fd, &c, 1)!=1)
			c=0;
	}
	else 
		c=sc->c, sc->c=0;
	return c;
}
static void unget(Scan *sc, char c) {
	sc->c=c;
}
static char *scan(Scan *sc) {
	static char str[1000];
	char *p=str;
	do {
		while((*p=get(sc)) && strchr(" \t\n\r", (*p)))
			;
		if(*p==';')
			while((*p=get(sc)) && !strchr("\n\r", (*p)))
				;
	}while(strchr(" \t\n\r", (*p)));
		
	if(isalpha(*p)) {
		while(isalnum((*(++p)=get(sc))))
			;
		unget(sc, *p); *p=0;
	}
	else if(isdigit(*p)) {
		while(isdigit((*(++p)=get(sc))))
			;
		unget(sc, *p); *p=0;
	}
	else
		*(p+1)=0;
	return str;
}
static int iscadr(char *p) {
	int i;
	for(i=1; p[i]=='a' || p[i]=='d'; i++)
		;
	return p[0]=='c' && p[i]=='r' && p[i+1]==0 && i>1;
}
static uint16_t _cadr(char *p, Scan *sc) {
	return 	*p=='a'? cons(sym.car, cons(_cadr(p+1, sc), nilAdrs)):
			*p=='d'? cons(sym.cdr, cons(_cadr(p+1, sc), nilAdrs)):
			*p=='r'? parse(0, sc): 0;
}
static void errQuit(char *msg, char *p) {
	fprintf(stderr, msg, p);
	exit(1);
}
static uint16_t parse(int depth, Scan *sc) {
	char *p=scan(sc);
	uint16_t a;
	if(*p==')' || *p==0)
		return nilAdrs;
	else if(iscadr(p)) {
		a=_cadr(p+1, sc);
		if(*(p=scan(sc))!=')')
			errQuit("syntax error: %s\n", p);
		return a;
	}
	else if(*p=='(')
		a=parse(depth+1, sc);
	else if(isalpha(*p))
		a=lookupSym(&sc->st, p);
	else if(isdigit(*p))
		a=lookupNum(&sc->nt, atoi(p));
	else {
		errQuit("syntax error: %s\n", p);
		return 0;
	}
	return depth>0? cons(a, parse(depth, sc)): a;
}
static void initSymTbl(SymTbl *p, uint16_t a, char **k) {
	p->sym=strdup(k[0]);
	p->adrs=((uint16_t*)&sym.nil)[a];
	p->next=(SymTbl*)calloc(1,sizeof(SymTbl));
	if(strcmp(k[0],"last")!=0) 
		initSymTbl(p->next, a+1, k+1);
}
static Scan* init(char *s) {
	Scan *sc=(Scan*)calloc(1,sizeof(Scan));
#ifdef ADMSERVER
	sc->fd = asock;
#else
	sc->fd=open(s,O_RDONLY);
#endif
	if(sc->fd<0) {
		fprintf(stderr,"can't open %s.\n", s);
		exit(1);
	}
	initSymTbl(&sc->st, 0, keyword);
	return sc;
}
static void finish(Scan *sc) {
	SymTbl *sp, *sq;
	NumTbl *np, *nq;
	for(sp=sc->st.next; sp; ) {
		sq=sp; sp=sp->next;
		free(sq->sym);
		free(sq);
	}
	for(np=sc->nt.next; np; ) {
		nq=np; np=np->next;
		free(nq);
	}
#ifndef ADMSERVER
	close(sc->fd);
#endif
	free(sc);
}

void execute(void *arg) {
	initThread();
	initVM();
	initData();
	initEnv();
	initComm();
	Scan *sc=init((char *)arg);
	//newGCRoot();
	uint16_t a=parse(0, sc);
	//printlist(a);
	eval(a, baseEnv, haltCont(newWord(hwId())));
	finish(sc);
}

