/**********************************************************
	rep.c
	version 1.1
	YoungMin Kwon (ykwon4@cs.uiuc.edu)
	Kirill Mechitov (mechitov@cs.uiuc.edu)
**********************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include "packet.h"

#define MaxClient	100

static void errquit(char *msg)
{
	fprintf(stderr, "error: %s\n", msg);
	exit(1);
}

void ttyInit(int fd)
{
	struct termios term;

	tcgetattr(fd, &term);
	term.c_lflag &= ~(ECHO | ICANON | IEXTEN | ISIG);
	term.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON);
	term.c_cflag &= ~(CSIZE | CSTOPB | PARENB);     /* NoParity, 8Bit, 1Stop */
	term.c_cflag |= CS8;
	term.c_oflag &= ~(OPOST);
	term.c_cc[VMIN]=36;
	term.c_cc[VTIME]=5;
	cfsetispeed(&term, B57600);                     /* speed: 57600bps */
	cfsetospeed(&term, B57600);
	tcsetattr(fd, TCSANOW, &term);
}

/*
void sendAck(int fd)
{
	Msg msg;
	uint16_t adrs=0xffff;
	msg.adrs[0]=adrs & 0xff;
	msg.adrs[1]=(adrs >> 8) & 0xff;
	msg.type=10;
	msg.group=0x7d;
	msg.length=1;
	msg.data[0]=TagAck;
	write(fd, &msg, PacketLen);
}

void recvAck(int fd)
{
	char buf[PacketLen];
	int i;
	for (;;) {
		for (read(fd, buf, 1); *buf != 0x7e; read(fd, buf, 1)) {
			printf("+");
		}
		for (i = 1; i < PacketLen; i += read(fd, buf + i, PacketLen - i) {
			printf("-");
		}
		if (((Msg*)buf)->data[0] == TagAck) {
			break;
		}
	}
	printf("ack ");
}
*/

void delay(long usec)
{
	struct timeval tv;
	tv.tv_sec = 0;
	tv.tv_usec = usec;
	select(1, NULL, NULL, NULL, &tv);
}

void run(char *serial, char *server, int port)
{
	int	sd, td;				/* socket descriptor, serial port tty descriptor */
	int	maxd;
	struct sockaddr_in sa;
	struct hostent *host;
	fd_set	rfd;

	if ((td = open(serial, O_RDWR)) < 0) {			/* serial port */
		errquit("open serial failed");
	}
	ttyInit(td);

	if ((host = gethostbyname(server)) == NULL) {	/* internet */
		errquit("gethostbyname failed");
	}

	memset(&sa, 0, sizeof (sa));
	sa.sin_family = AF_INET;
	memcpy(&sa.sin_addr, host->h_addr, host->h_length);
	sa.sin_port = htons(port);

	if ((sd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		errquit("socket failed");
	}
	if (connect(sd, (struct sockaddr*) &sa, sizeof (sa)) < 0) {
		errquit("connect failed");
	}

	maxd = sd > td ? sd: td;

	for (;;) {
		char buf[PacketLength];
		Msg *msg = (Msg*)buf;

		FD_ZERO(&rfd);
		FD_SET(sd, &rfd);
		FD_SET(td, &rfd);

		if (select(maxd + 1, &rfd, NULL, NULL, NULL) < 0) {
			errquit("client select failed");
		}

		if (FD_ISSET(td, &rfd)) {
			int i, j;
			for (read(td, buf, 1); *buf != 0x7e; read(td, buf, 1)) {
				fprintf(stderr, "!");
			}

			/* with crc */
			for (i = 1; i < PacketLength; i += j) {
				if ((j = read(td, buf + i, PacketLength - i)) < 0) {
					errquit("client read failed");
				}
			}
			buf[0] = buf[1] = (char)0xff;
			fprintf(stderr, "t:%d ", msg->length);

			/* without crc */
			for (i = 0; i < msg->length + PacketHeaderLength; i += j) {
				if ((j = write(sd, buf + i, msg->length + PacketHeaderLength - i)) < 0) {
					errquit("client write failed");
				}
			}
		}
		else if (FD_ISSET(sd, &rfd)) {
			int	i, j;
			for (i = 0; i < PacketHeaderLength; i += j) {
				if ((j = read(sd, buf + i, PacketHeaderLength - i)) < 0) {
					errquit("client read failed");
				}
			}

			for (i = 0; i < msg->length; i += j) {
				if ((j = read(sd, buf + i + PacketHeaderLength, msg->length - i)) < 0) {
					errquit("client read failed");
				}
			}
			fprintf(stderr, "s:%d ", msg->length);

			for (i = j = 0; i < PacketLength; i += j) {
				if ((j = write(td, buf + i, PacketLength - j)) < 0) {
					errquit("client write");
				}
				/*
				delay(500*1000);
				delay(300*1000);
				*/
				delay(300*1000);
			}
		}
	}
}

int main(int argc, char **argv)
{
	if (argc != 4) {
		fprintf(stderr, "usage: %s <serialdev> <host> <port>\n", argv[0]);
		exit(0);
	}
	run(argv[1], argv[2], atoi(argv[3]));	
	return 0;
}
