/**********************************************************
	comm.c
	version 1.1
	YoungMin Kwon (ykwon4@cs.uiuc.edu)
	version 1.2
	Sergei Shevlyagin (shevlyag@uiuc.edu)
**********************************************************/
#ifdef PC
#include <stdio.h>
#endif
#include <string.h>
#include "data.h"
#include "memory.h"
#include "eval.h"
#include "comm.h"
#include "hw.h"
#include "thread.h"
#include "extcomm.h"

#define SelfCmdSendCons	0xffff

uint16_t comAdrs;
static ComData *comData;
static RawBufQ rawBufQ;


// keep track of last message sent in stop-and-go
lastMessage lm_tx[NumRawBuf];
lastMessage lm_rx[NumRawBuf];


#ifdef PC
pthread_mutex_t rx_mt, rb_mt;
uint8_t timer_flag;
#endif

void initComm()
{
	// default initilization
	ComData *p;
	uint8_t i;
	comAdrs=newBytes(sizeof(ComData));	
	p=(ComData*)(fetch(comAdrs, 1)+sizeof(Bytes));			//p is volatile
	p->sbi=3, p->sb[2]=0;
	p->rbi=0, p->rbn=0, p->rbnext=0;
	for(i=0; i<NumStream; i++) {
		p->lru[i]=i;
		memset(&(p->stream[i]), 0, sizeof(StreamHead));
	}

#ifdef PC	
	// throttle timer
	setup_timer_throttle();
#endif	
	rawBufQ.h=rawBufQ.n=0;
	// set up for transmission of first message
	//ycylm_tx[rawBufQ.h].ok = 1; 	
	initSP(sendPacket);
}

uint8_t isRawBufFull()
{
	return rawBufQ.n==NumRawBuf;
}

uint8_t isRawBufEmpty()
{
	return rawBufQ.n==0;
}



void addRawBuf(uint8_t *data, uint8_t n)
{
	acknowledgement ack, rAck; 
	uint8_t i; 
	uint8_t repeat =0; 
	if(isRawBufFull()) {
		ledSet(7);
	}
	// check if Ack
	if(data[2] == 0xff)
	{
		memcpy(&rAck, data, 6);
		for(i = 0; i<8; i++)
		{
			if(lm_tx[i].senderId == rAck.txId && lm_tx[i].seqNo == rAck.txSeqNo)
			{
				lm_tx[i].ok = 1; 
				break; 
			}
		}
		
	}
	else
	{
		// send Ack
		memcpy(&ack.txId, data, 2);
		ack.rxId = hwId();
		ack.rxSeqNo = 0xff; 
		ack.txSeqNo = data[2];
#ifdef PC
		if(!message_troll())
		{
			sendPacket((uint8_t*)&ack,6);
			//debugAck("Send Ack", ack);
		}//debugAck(ack);
#else
		sendPacket((uint8_t*)&ack,6);
#endif
		// if not a repeat add to raw buffer
		for(i = 0; i<8; i++)
		{
			if(lm_rx[i].senderId == ack.txId && lm_rx[i].seqNo == ack.txSeqNo)
			{
#ifdef DEBUG
				debugPrint("double pakcet", data, n);
#endif
				repeat = 1;
				break; 
			}
		}
		if(!repeat)
		{
			lm_rx[rawBufQ.h].senderId = ack.txId;
			lm_rx[rawBufQ.h].seqNo = ack.txSeqNo;
			
		
			memcpy(&rawBufQ.rawBuf[rawBufQ.h].senderId, data, n);
#ifdef DEBUG		
			debugPrint("addRawBuf", data, n);
#endif
				
			rawBufQ.rawBuf[rawBufQ.h].dataLen=n-3;					//3: senderId, seqNo
			rawBufQ.h=(rawBufQ.h+1)%NumRawBuf;
			rawBufQ.n++;
		}
	}
}


RawBuf* delRawBuf()
{
	return rawBufQ.n==0? (RawBuf *)0: &rawBufQ.rawBuf[(rawBufQ.h+NumRawBuf-rawBufQ.n--)%NumRawBuf];
}

static uint16_t revStreamBuf(uint16_t self, uint16_t prev)			//stacked -> queued
{
	uint16_t start;
	StreamBuf *p=(StreamBuf*)(fetch(self, 0)+sizeof(Bytes));		//p is volatile
	if(p->next==0) {
		p=(StreamBuf*)(fetch(self, 1)+sizeof(Bytes));				//to make it dirty
		p->next=prev;
		return self;
	}
	else {
		start=revStreamBuf(p->next, self);
		p=(StreamBuf*)(fetch(self, 1)+sizeof(Bytes));				//p is volatile
		p->next=prev;
		return start;
	}
}

void saveStream()
{
	RawBuf *rb;
	uint8_t seq, h, i, j;
	ComData *p;
	StreamBuf *q;
#ifdef PC
	pthread_mutex_lock(&rb_mt);
#endif
	p=(ComData*)(lock(comAdrs, 1)+sizeof(Bytes));	
	
	while(!isRawBufEmpty()) {
		rb=delRawBuf();
		if(rb->dataLen==0) {
			continue;
		}
    	
	    for(i=0; i<NumStream; i++) {						//lookup stream
			if(p->stream[p->lru[i]].sender==rb->senderId) {
				break;
			}
		}
		if(i==NumStream) {									//kick out lru buffer
			i=NumStream-1;
		}
		h=p->lru[i];
    	
		seq=rb->seqNo & 0x7f;
		if(seq==0) {										//new stream
			p->stream[h].sender=rb->senderId;
			p->stream[h].seqNo=seq;
			p->stream[h].buffer=newBytes(rb->dataLen+2);	//2:next link address
			q=(StreamBuf*)(fetch(p->stream[h].buffer, 1)+sizeof(Bytes));	//q is volatile
			q->next=0;
			memcpy(&(q->dataStart), rb->data, rb->dataLen);		
		}
		else if(p->stream[h].seqNo+1==seq) {				//existing stream
			uint16_t a=newBytes(rb->dataLen+2);				//2:next link address
			q=(StreamBuf*)(fetch(a, 1)+sizeof(Bytes));		//q is volatile
			q->next=p->stream[h].buffer;
			memcpy(&(q->dataStart), rb->data, rb->dataLen);
			p->stream[h].seqNo=seq;
			p->stream[h].buffer=a;
		}
		else { 										//neither new nor existing stream
			ledSet(7);
			break;
		}
		for(j=i; j>0; j--) {							//update lru table
			p->lru[j]=p->lru[j-1];
		}
		p->lru[0]=h;
		if(rb->seqNo & 0x80) {						//last packet
			uint8_t *r=fetch(revStreamBuf(p->stream[h].buffer, 0), 0);
			for(j=1; j<NumStream; j++) {			//h is lru now
				p->lru[j-1]=p->lru[j];
			}
			p->lru[j-1]=h;
			p->stream[h].buffer=0;					//if it is not zero, gc will do something
			p->rbi=0;								//rb[26]
			p->rbn=*(uint16_t*)(r+1)-2;				//r (Bytes): tag, size, next, data
			p->rbnext=*(uint16_t*)(r+3);
			memcpy(p->rb, r+5, p->rbn);
			deliverMsg(decode());
			wakeupThreads();
		}
	}
	
	unlock(comAdrs, 1);
#ifdef PC
	pthread_mutex_unlock(&rb_mt);
#endif
}


// connection to send
static void flush(uint8_t isLast)			// while sending a packet may be received, but not vice versa
{
	while(isSendPending()) {		// check this part//////////////////////////////////////////////
		yield();
	}
	if(isLast) {
		comData->sb[2] |= 0x80;
	}
	sendPacketAck(comData->sb, comData->sbi, lm_tx, rawBufQ.h);
	comData->sb[2]=isLast? 0: comData->sb[2]+1;		//sequence number
	comData->sbi=3;
}

static void send(void *_p, uint8_t n)
{
	uint8_t *p=_p, i;
	for(i=0; i<n; i++) {
		if(comData->sbi>=sizeof(comData->sb))  {
			flush(0);
		}
		comData->sb[comData->sbi++]=p[i];
	}
}

static void recv(void *_p, uint8_t n)				//read n bytes from p->rb
{
	uint8_t *p=_p, i; 
	for(i=0; i<n; i++) {
		if(comData->rbi>=comData->rbn) {
			uint8_t *r=fetch(comData->rbnext, 0);
#ifdef DEBUG
			if(comData->rbnext==0) {
				fprintf(stderr,"error. recv: need more than rececived\n");
			}
#endif
			comData->rbi=0;								//rb[26]
			comData->rbn=*(uint16_t*)(r+1)-2;			//r (Bytes): tag, size, next, data
			comData->rbnext=*(uint16_t*)(r+3);
			memcpy(comData->rb, r+5, comData->rbn);
		}
		p[i]=comData->rb[comData->rbi++];
	}
}

static void sendByte(uint8_t b)
{
	send(&b, 1);
}

static void sendWord(uint16_t w)
{
	send(&w, sizeof(uint16_t));
}

static void sendData(uint8_t *p)
{
	uint8_t q=p[0];
	p[0] &= 0x0f;
	switch(p[0]) {
		case TagSymbol:	send(p, sizeof(Symbol)); break;
		case TagByte:	send(p, sizeof(Byte)); break;
		case TagWord:	send(p, sizeof(Word)); break;
		case TagDWord:	send(p, sizeof(DWord)); break;
		case TagFloat:	send(p, sizeof(Float)); break;
#ifdef DEBUG
		default:
			fprintf(stderr,"error. sendData: illigal tag %d to send.", p[0]);
#endif
	}
	p[0]=q;
}

static uint8_t recvByte()
{
	uint8_t b;
	recv(&b, 1);
	return b;
}

static uint16_t recvWord()
{
	uint16_t w;
	recv(&w, sizeof(uint16_t));
	return w;
}

static uint16_t recvData(uint8_t tag)
{
	if (tag == TagSymbol) {
		uint16_t data; recv(&data, sizeof(uint16_t));
		return newSymbol(data);
	}
	else if (tag == TagByte) {
		uint8_t data; recv(&data, sizeof(uint8_t));
		return newByte(data);
	}
	else if (tag == TagWord) {
		uint16_t data; recv(&data, sizeof(uint16_t));
		return newWord(data);
	}
	else if (tag == TagDWord) {
		uint32_t data; recv(&data, sizeof(uint32_t));
		return newDWord(data);
	}
	else if (tag == TagFloat) {
		float data; recv(&data, sizeof(float));
		return newFloat(data);
	}
	else {
		return 0;
	}
}

static uint16_t *getRef(uint16_t inx, uint8_t rw)				//index to Ref*
{
	uint16_t row=inx/NumCol, col=inx%NumCol;
	if(comData->ref[row]==0) {
		comData->ref[row]=newBytes(NumCol*sizeof(uint16_t));
		memset(fetch(comData->ref[row],1)+sizeof(Bytes), 0, NumCol*sizeof(uint16_t));
	}
	return (uint16_t*)(fetch(comData->ref[row], rw)+sizeof(Bytes)+col*sizeof(uint16_t));
}

static uint16_t _findRef(uint16_t a)		//return the index of an ref with ref.a==a
{
	uint16_t i;
	for(i=0; i<comData->nref; i++) {
		if(*getRef(i, 0)==a) {
			return i;
		}
	}
	return 0xffff;
}

static void _addRef(uint16_t a)
{
	if(_findRef(a)==0xffff) {	//if not already added
#ifdef DEBUG
		if(comData->nref>=NumRow*NumCol) {
			fprintf(stderr, "error: _addRef\n");
		}
#endif
		*getRef(comData->nref, 1)=a;	//set current reference to a
		comData->nref++;
	}
}

static void _mark(uint16_t a)
{
	AdrsStack astack;
	uint16_t buf[100];
	uint8_t *p;
	initStack(&astack, buf, 100);
	push(&astack, a);
	while(!isEmptyStack(&astack)) {
		a=pop(&astack);
#ifdef DEBUG
		if(!isReserved(a)) {
			fprintf(stderr, "error. _mark: unreserved memory access %d.\n", a);
		}
#endif
		//if(a<=baseEnv) : error because there are gaps in reserved memory below baseEnv
		if(a<=sym.last || a==baseEnv) {
			continue;
		}
		p=fetch(a,1);
		switch(p[0] & 0x2f) {
			case TagSymbol:
			case TagByte:
			case TagWord:
			case TagDWord:
			case TagFloat:
				*p|=0x20;
				break;
			case TagPair:
				*p|=0x20;
				push(&astack, ((Pair*)p)->second);
				push(&astack, ((Pair*)p)->first);
				break;
			default: 
				if(p[0]&0x20) {
					_addRef(a);	//add new reference if it is not already saved
				}
		}
	}
}

static void _unmark(uint16_t a)
{
	AdrsStack astack;
	uint16_t buf[100], i;
	uint8_t *p;
	initStack(&astack, buf, 100);
	push(&astack, a);
	while(!isEmptyStack(&astack)) {
		a=pop(&astack);
		if(a==SelfCmdSendCons) {
			sendByte(CmdCons);
			continue;
		}
#ifdef DEBUG
		if(!isReserved(a)) {
			fprintf(stderr, "error. _unmark: unreserved memory access %d.\n", a);
		}
#endif
		//if(a<=baseEnv) : error because there are gaps in reserved memory below baseEnv
		if(a<=sym.last || a==baseEnv) {
			sendByte(TagAdrs), sendWord(a);
			continue;
		}
		p=fetch(a,1);
		switch((p[0] & 0x2f)^0x20) {
			case TagSymbol:	
			case TagByte:	
			case TagWord:
			case TagDWord:
			case TagFloat:
				*p^=0x20;
				sendData(p);
				if((i=_findRef(a))!=0xffff) {
					sendByte(CmdSaveRef), sendWord(i);
				}
				break;
			case TagPair:
				*p^=0x20;
				push(&astack, SelfCmdSendCons);
				push(&astack, ((Pair*)p)->second);	//p is volatile
				push(&astack, ((Pair*)p)->first);	
				sendByte(TagPair);
				if((i=_findRef(a))!=0xffff) {
					sendByte(CmdSaveRef), sendWord(i);
				}
				break;
			default: 
				if(!(p[0]&0x20)) {
					sendByte(TagRef), sendWord(_findRef(a));
				}
		}
	}
}

static uint16_t _decode()
{
	AdrsStack astack;
	uint16_t buf[100], r;
	uint8_t cmd;
	initStack(&astack, buf, 100);
	while((cmd=recvByte())!=CmdEnd) {
		switch(cmd) {
			case TagSymbol:	
			case TagByte:	
			case TagWord:
			case TagDWord:
			case TagFloat:
				push(&astack, recvData(cmd));
				break;
			case TagPair: 		//save place holder
				push(&astack, cons(nilAdrs, nilAdrs));
				break;
			case TagAdrs:		//adress in baseEnv
				push(&astack, recvWord());
				break;
			case TagRef:
				push(&astack, *getRef(recvWord(), 0));
				break;
			case CmdSaveRef:
				*getRef(recvWord(), 1)=top(&astack);
				break;
			case CmdCons: {
				uint16_t a=pop(&astack), b=pop(&astack), c=top(&astack);
				setcar(c, b);
				setcdr(c, a);
				break;
			}
#ifdef DEBUG				
			default:
				fprintf(stderr,"error. decode: unknown cmd %d.\n", cmd);
#endif
		}
	}
	r=pop(&astack);
#ifdef DEBUG
	if(!isEmptyStack(&astack)) {
		fprintf(stderr,"error. decode cmd %d: astack is not empty.\n", cmd);
	}
#endif
	return r;
}

void encode(uint16_t a, uint16_t id)
{
	uint8_t i;
	comData=(ComData*)(lock(comAdrs, 1)+sizeof(Bytes));
	//printlist(a);
	comData->nref=0;
	*((uint16_t*)comData->sb)=id;	//sending actor id
	comData->sb[2]=0;				//sequence number
	for(i=0; i<NumRow; i++) {
		comData->ref[i]=0;
	}
	_mark(a);
	_unmark(a);
	sendByte(CmdEnd);
	flush(1);
	unlock(comAdrs, 1);
}

uint16_t decode()
{
	uint8_t i;
	uint16_t a;
	comData=(ComData*)(lock(comAdrs, 1)+sizeof(Bytes));
	comData->nref=0;
	for(i=0; i<NumRow; i++) {
		comData->ref[i]=0;
	}
	a=_decode();
	unlock(comAdrs, 1);
	//printlist(a);
	return a;
}

uint16_t deliverMsg(uint16_t msg)	//msg (adrs, content): deliver msg to right thread
{
	uint16_t rid=*((uint16_t*)deref(car(msg))), id, msgq;
	uint8_t i;
	if(rid==0) {					//broadcast -> launcher
		rid=hwId();
	}
	for(i=thr.t; i!=thr.h; i=thr.next(i)) {
		id=*((uint16_t*)deref(getId(thr.cont[i])));
		if(id==rid) {
			msgq=getMsgQ(thr.cont[i]);
#ifdef DEBUG
			if(msgq==nilAdrs) {
				fprintf(stderr,"error. deliverMsg: cannot find msgq.\n");
			}
#endif
			setcdr(msgq, cons(msg, cdr(msgq)));
			break;
		}
	}
	return msg;
}
