/**********************************************************
	comm.h
	version 1.1
	YoungMin Kwon (ykwon4@cs.uiuc.edu)
	version 1.2
	Sergei Shevlyagin (shevlyag@uiuc.edu)
**********************************************************/
#ifndef _COMM_H
#define _COMM_H

#include "memory.h"
#define NumStream	2				//maximum number of streams that can be received concurrently
//#define NumRawBuf	4				//maximum number of raw buffers
#define NumRawBuf	8				//maximum number of raw buffers
#define NumCol  25					//used in encode/decode algorithm to store references
#define NumRow  9					//used in encode/decode algorithm to store references

typedef struct {
	uint16_t sender;				//sender id
	uint8_t seqNo;					//sequence number
	uint16_t buffer;				//stream buffer, first two bytes of buffer is a link to the next buffer
} StreamHead;
typedef struct {
	uint16_t next;					//next buffer address
	uint8_t dataStart;				//&dataStart: beginning of data array
} StreamBuf;
typedef struct {
	uint8_t	dataLen;
	uint16_t senderId;
	uint8_t seqNo;
	uint8_t data[29-3];				//3: senderId, seqNo
} RawBuf;
typedef struct {					//circular queue of RawBuffers
	uint8_t	h:4;					//head: where to put new data
	uint8_t n:4;					//number of entries in Queue
	RawBuf rawBuf[NumRawBuf];
} RawBufQ;
typedef struct {
	uint8_t sb[29], sbi;			//sb[0..1]: sender id, sb[2]: seqNo, sb[3..]: data
	uint8_t	rb[26], rbi, rbn;		//for decode
	uint16_t rbnext;
	StreamHead stream[NumStream];
	uint8_t lru[NumStream];			//sorted index to stream

	uint16_t ref[NumRow];		
	uint16_t nref;
} ComData;
extern uint16_t comAdrs;

void initComm();
uint8_t isRawBufFull();
uint8_t isRawBufEmpty();
void addRawBuf(uint8_t *data, uint8_t n);
RawBuf* delRawBuf();
void saveStream();
void encode(uint16_t a, uint16_t id);	//send data a, id: sendder id, receiver is encoded in message
uint16_t decode();						//receive data
uint16_t deliverMsg(uint16_t msg);		//deliver it to correct thread

#endif /* _COMM_H */
