/**********************************************************
	data.c
	version 1.1
	YoungMin Kwon (ykwon4@cs.uiuc.edu)
**********************************************************/
#ifdef PC
#include <stdio.h>
#endif
#include <string.h>
#include <stdarg.h>
#include "memory.h"
#include "data.h"
#include "eval.h"
#include "thread.h"
#include "comm.h"
#include "hw.h"

#define MaxGCRoot	20
static uint16_t gcRoot[MaxGCRoot+1];	//linked list begins from gcRoot[MaxGCRoot]
static uint16_t gci=0;
static uint16_t _reserve(uint16_t n);
static uint8_t mask, nextMask=0x80;
static uint16_t sweepAdrs;
static AdrsStack astack;

void initData() {
	mask=0, nextMask=0x80, sweepAdrs=0;
	gci=0;
}
#define newInstance(type, name)			\
uint16_t new##name(type value) {		\
	uint16_t a=_reserve(sizeof(name));	\
	name *p=(name*)fetch(a,1);			\
	p->tag=(Tag##name & 0x3f) | mask;	\
	p->data=value;						\
	return a;							\
}
uint16_t newPair(uint16_t a, uint16_t b) {
	uint16_t c=_reserve(sizeof(Pair));
	Pair *p=(Pair *)fetch(c,1);
	p->tag=(TagPair & 0x3f) | mask;
	p->first=a, p->second=b;
	return c;
}
uint16_t newBytes(uint16_t n) {
	uint16_t a=_reserve(n+sizeof(Bytes));
	Bytes *p=(Bytes *)fetch(a,1);
	p->tag=(TagBytes & 0x3f) | mask;
	p->size=n;
	return a;
}
newInstance(uint16_t, Symbol)
newInstance(uint8_t, Byte)
newInstance(uint16_t, Word)
newInstance(uint32_t, DWord)
newInstance(float, Float)
uint16_t name(char *str) {
	//return newSymbol(((int)str) & 0xffff);
	uint16_t a=reserve(sizeof(Symbol));
	Symbol *p=(Symbol*)fetch(a,1);
	p->tag=(TagSymbol & 0x3f) | mask;
	p->data=((int)str+baseEnv) & 0xffff;
#ifdef DEBUG
	// FIXME
	/*
	if(p->data<=baseEnv && baseEnv>0)
		fprintf(stderr,"error. name: name hash collision %d.\n", p->data);
	*/
#endif
	return a;
}
uint16_t cons(uint16_t a, uint16_t b) {
#ifdef DEBUG
	uint8_t p=*fetch(a, 0) & 0x0f;
	uint8_t q=*fetch(b, 0) & 0x0f;
	if(!(TagSymbol <= p && p <= TagBytes) || a>MaxAddress)
		fprintf(stderr, "error: cons_a: %d, %d, %d\n", p, a, isReserved(a));
	if(!(TagSymbol <= q && q <= TagBytes) || b>MaxAddress)
		fprintf(stderr, "error: cons_b: %d, %d, %d\n", q, b, isReserved(b));
#endif
	return newPair(a, b);
}
uint16_t car(uint16_t a) {
	Pair *p=(Pair*)fetch(a, 0);
#ifdef DEBUG
	if((p->tag & 0x0f)!=TagPair)
		fprintf(stderr, "\nerror: car: %d, %d, %p, %d\n", p->tag, a, p, isReserved(a));
#endif
	return p->first;
}
uint16_t cdr(uint16_t a) {
	Pair *p=(Pair*)fetch(a, 0);
#ifdef DEBUG
	if((p->tag & 0x0f)!=TagPair)
		fprintf(stderr, "\nerror: cdr: %d, %d, %p, %d\n", p->tag, a, p, isReserved(a));
#endif
	return p->second;
}
static uint8_t sizeofData(uint8_t *data) {
	uint8_t tag=*data & 0x0f;
	return	tag==TagSymbol? sizeof(Symbol):
			tag==TagByte?	sizeof(Byte):
			tag==TagWord?	sizeof(Word):
			tag==TagDWord?	sizeof(DWord):
			tag==TagFloat?	sizeof(Float):
			tag==TagPair?	sizeof(Pair):
			tag==TagBytes?	sizeof(Bytes)+*((uint16_t*)(data+1)): 0;
}
uint16_t setdata(uint16_t dst, uint16_t src) {
	uint8_t *p=fetch(dst, 1);
	uint8_t *q=fetch(src, 1);
#ifdef DEBUG
	if((*p & 0x0f)!=(*q & 0x0f))
		fprintf(stderr, "error: setdata, type mismatch\n");
#endif
	memcpy(p, q, sizeofData(q));
	return dst;
}
uint16_t setcar(uint16_t a, uint16_t v) {
	Pair *p=(Pair*)fetch(a, 1);
#ifdef DEBUG
	if((p->tag & 0x0f)!=TagPair || v>MaxAddress)
		fprintf(stderr, "error: setcar\n");
#endif
	return p->first=v;
}
uint16_t setcdr(uint16_t a, uint16_t v) {
	Pair *p=(Pair*)fetch(a, 1);
#ifdef DEBUG
	if((p->tag & 0x0f)!=TagPair || v>MaxAddress)
		fprintf(stderr, "error: setcdr\n");
#endif
	return p->second=v;
}
uint16_t listlen(uint16_t c) {
	return c==nilAdrs? 0: 1+listlen(cdr(c));
}
static uint16_t _list(int n, va_list ap) {
	if(n>0) {
		uint16_t a=va_arg(ap, int);
		return cons(a, _list(n-1, ap));
	}
	else
		return nilAdrs;
}
uint16_t list(uint16_t n, ...) {
	va_list ap;
	uint16_t a;
	va_start(ap, n);
	a=_list(n, ap);
	va_end(ap);
	return a;
}
uint16_t quote(uint16_t a) {
	return list(2, sym.quote, a);
}

void initStack(AdrsStack *p, uint16_t *_as, uint16_t _na) {
	p->as=_as, p->na=_na, p->sp=0;
}
void push(AdrsStack *p, uint16_t a) {
#ifdef DEBUG
	if(p->sp>=p->na)
		fprintf(stderr,"astack overflow\n");
#endif
	p->as[p->sp++]=a;
}
uint16_t pop(AdrsStack *p) {
#ifdef DEBUG
	if(p->sp<=0)
		fprintf(stderr,"astack underflow\n");
#endif
	return p->as[--p->sp];
}
uint16_t top(AdrsStack *p) {
#ifdef DEBUG
	if(p->sp<=0)
		fprintf(stderr,"astack underflow\n");
#endif
	return p->as[p->sp-1];
}
uint8_t isEmptyStack(AdrsStack *p) {
	return p->sp==0;
}
uint8_t isFullStack(AdrsStack *p) {
	return p->sp>=p->na;
}
static void _mark(uint16_t a) {
	uint16_t buf[100];
#ifdef STAT
	int	dgcwr=-stat.wr;
#endif
	mask=nextMask;
	initStack(&astack, buf, 100);
	push(&astack, a);
	while(!isEmptyStack(&astack)) {
		uint8_t *p=fetch(pop(&astack), 1);
		if(p[0] & mask)
			continue;
		switch(p[0] & 0x0f) {		
			case TagSymbol:
			case TagByte:
			case TagWord:
			case TagDWord:
			case TagFloat:	
			case TagBytes:	*p = (*p & 0x3f) | mask; break;
			case TagPair: {
				*p = (*p & 0x3f) | mask;
				push(&astack, ((Pair*)p)->second);
				push(&astack, ((Pair*)p)->first);
				break;
			}
		}
	}
#ifdef STAT
	dgcwr+=stat.wr;
	stat.markwr+=dgcwr;
	stat.nmark++;
#endif
}
static uint32_t _sweep() {
	uint16_t n, c=0;
#ifdef STAT
	int i, j, loop;
#else
	uint8_t i, j, loop;
#endif
	if(sweepAdrs<baseEnv)				//skip gc up to baseEnv
		sweepAdrs=(baseEnv/PageSize)*PageSize;
	reserveStartPage=sweepAdrs/PageSize;	//search begins from the reserveStartPage for reservation
	for(loop=0; loop<10 && sweepAdrs<MaxAddress; loop++, sweepAdrs+=PageSize) {
		Page *page=(Page*)fetch(sweepAdrs, 0);
		for(i=0; i<PageSize; ) {
			if(sweepAdrs+i<baseEnv) {	//skip gc up to baseEnv
				i=baseEnv-sweepAdrs;
				continue;
			}
			if(page->bitmap[i/8] & (1<<(i%8))) {//if reserved
				uint8_t *p=&page->memory[i];
#ifdef DEBUG
				uint8_t tag=*p & 0x0f;			
/*
					n =	tag==TagSymbol?	sizeof(Symbol):
					tag==TagByte?	sizeof(Byte):
					tag==TagWord?	sizeof(Word):
					tag==TagDWord?	sizeof(DWord):
					tag==TagFloat?	sizeof(Float):
					tag==TagPair?	sizeof(Pair): 
					tag==TagBytes?	sizeof(Bytes)+*(uint16_t*)(p+1): 0;
*/
#endif
				n=sizeofData(p);
#ifdef DEBUG
				if(n==0)
					fprintf(stderr, "error. gc: %d:%d ", sweepAdrs, tag);
#endif
				if((*p & mask)==0) {
					for(j=i; j<i+n; j++)
						page->bitmap[j/8] &= (0xff ^ (1<<(j%8)));	//release
					page->bits.flag.dirty=1;
					c+=n;
				}
				i+=n;
			}
			else {
				i++;
			}
		}
	}
	if(sweepAdrs>=MaxAddress) {
		mask=0, nextMask ^= 0xc0, sweepAdrs=0;
	}
	reserveEndPage=sweepAdrs/PageSize;	//search ends at the reserveEndPage for reservation
	return c;
}
void addGCRoot(uint16_t a) {
	if(gci<MaxGCRoot)
		gcRoot[gci++]=a;
}
void newGCRoot() {
	gci=0;
	gcRoot[gci]=nilAdrs;
}
uint16_t gc() {		//return number of uint8_ts released
	uint16_t i, j, a;
	extern uint16_t comAdrs;		//communication buffer
#ifdef STAT
	int	dgcwr=-stat.wr;	
	stat.ngc++;
#endif
	if(mask==0) {					//in this case mark otherwise just sweep
		ComData *p=(ComData*)(lock(comAdrs, 1)+sizeof(Bytes));
		_mark(comAdrs);
		for(i=0; i<NumStream; i++) {
			for(a=p->stream[i].buffer; a!=0; ) {
				_mark(a);
				a=((StreamBuf*)(fetch(a, 0)+sizeof(Bytes)))->next;
			}
		}
		unlock(comAdrs, 1);
		j=gci<MaxGCRoot? gci: MaxGCRoot+1;
		for(i=0; i<j; i++)			//temporary memory of current continuation
			_mark(gcRoot[i]);
		for(i=thr.t; i!=thr.h; i=thr.next(i)) {
			_mark(thr.cont[i]);
			_mark(thr.value[i]);
		}
	}
	i=_sweep();
#ifdef STAT
		dgcwr+=stat.wr;
		stat.gcwr+=dgcwr;
		stat.maxgcwr=stat.maxgcwr>dgcwr? stat.maxgcwr: dgcwr;
#endif
	return i;
}
static uint16_t _reserve(uint16_t n) {
	uint16_t m=gci<MaxGCRoot? n: sizeof(Pair)+n;
	uint16_t a=reserve(m);
	while(a==0) {
#ifdef STAT
		stat.regc++;	//gc again
#endif
		gc();
		//printf("gc: %d\n", gc());
		a=reserve(m);
#ifdef DEBUG
	// FIXME
	/*
		if(a==0) 
			fprintf(stderr,"reserve: not enough memory for %d bytes.\n", n);
	*/
#endif
	}
	if(gci<MaxGCRoot) {
		addGCRoot(a);
		return a;
	}
	else {
		Pair *p=(Pair*)fetch(a,1);
		p->tag=(TagPair & 0x3f) | mask; p->first=a+sizeof(Pair); p->second=gcRoot[gci];
		gcRoot[gci]=a;
		return p->first;
	}
}

