/**********************************************************
	data.h
	version 1.1
	YoungMin Kwon (ykwon4@cs.uiuc.edu)
**********************************************************/
#ifndef _DATA_H
#define _DATA_H

#include "memory.h"
typedef enum {
	TagSymbol=1, TagByte, TagWord, TagDWord, TagFloat, TagPair, TagBytes
} Tags;
typedef enum {
	TagAdrs=TagBytes+1, TagRef, CmdSaveRef, CmdCons, CmdEnd
//	CmdSaveRef=TagBytes+1, CmdRef, CmdPreparePair, CmdSetPair, CmdBaseRef, CmdEnd
} Cmds;
typedef struct {
	uint8_t tag;
	uint16_t data;
} Symbol;
typedef struct {
	uint8_t tag;
	uint8_t data;
} Byte;
typedef struct {
	uint8_t tag;
	uint16_t data;
} Word;
typedef struct {
	uint8_t tag;
	uint32_t data;
} DWord;
typedef struct {
	uint8_t tag;
	float data;
} Float;
typedef struct {
	uint8_t tag;
	uint16_t first, second;
} Pair;
typedef struct {
	uint8_t tag;
	uint16_t size;		//current impl: size <= 112
} Bytes;

typedef struct {
	uint16_t *as, na, sp;
} AdrsStack;
#define cadr(a)		car(cdr(a))
#define caddr(a)	cadr(cdr(a))
#define cadddr(a)	caddr(cdr(a))
#define caddddr(a)	cadddr(cdr(a))
#define caar(a)		car(car(a))
#define cdar(a)		cdr(car(a))
#define cddr(a)		cdr(cdr(a))
#define cdddr(a)	cdr(cddr(a))
#define cddddr(a)	cdr(cdddr(a))

void initData();
uint16_t newSymbol(uint16_t v);
uint16_t newByte(uint8_t v);
uint16_t newWord(uint16_t v);
uint16_t newDWord(uint32_t v);
uint16_t newFloat(float v);
uint16_t newPair(uint16_t a, uint16_t b);
uint16_t newBytes(uint16_t n);
uint16_t name(char *str);
uint16_t cons(uint16_t a, uint16_t b);
uint16_t car(uint16_t a);
uint16_t cdr(uint16_t a);
uint16_t setdata(uint16_t dst, uint16_t src);
uint16_t setcar(uint16_t a, uint16_t v);
uint16_t setcdr(uint16_t a, uint16_t v);
uint16_t listlen(uint16_t a);
uint16_t list(uint16_t n, ...);
uint16_t quote(uint16_t a);
uint16_t gc();
void addGCRoot(uint16_t a);
void newGCRoot();
void initStack(AdrsStack *p, uint16_t *as, uint16_t na);
void push(AdrsStack *p, uint16_t a);
uint16_t pop(AdrsStack *p);
uint16_t top(AdrsStack *p);
uint8_t isEmptyStack(AdrsStack *p);
uint8_t isFullStack(AdrsStack *p);

#endif /* _DATA_H */
