/**********************************************************
	eval.c
	version 1.1
	YoungMin Kwon (ykwon4@cs.uiuc.edu)
	Kirill Mechitov (mechitov@cs.uiuc.edu)
**********************************************************/
#ifdef PC
#include <stdio.h>
#include <stdlib.h>
#endif
#include <stdarg.h>
#include "memory.h"
#include "data.h"
#include "eval.h"
#include "comm.h"
#include "thread.h"
#include "hw.h"
#include "prim.h"

#ifdef PC
char *keyword[]={
	"nil",		"true",		"false",
	"lambda",	"rec",		"cond",		"halt",		"cond_",	"apply",	"closure",	"quote",
	"add",		"sub",		"mul",		"div",
	"equal",	"gt",	 	"ge", 		"lt",	 	"le",
	"and", 		"or", 		"not",
	"cons", 	"car", 		"cdr", 		"list",
	"set",	"setcar", 	"setcdr",
	"io", 		"print",	"send",		"id",		"msgq",
	"sleep",
	"eval",		"callcc",	"seq",		"par",
	"prim", "extmsg",
	"last"
};
#endif
Symbols sym;
uint16_t baseEnv, nilAdrs;
static uint16_t thrId=1;	//thread id

uint16_t extendEnv(uint16_t env, uint16_t ids, uint16_t values) {
	return ids==nilAdrs? env:
		extendEnv(cons(cons(car(ids), car(values)), env), cdr(ids), cdr(values));
}
void *deref(uint16_t a) {
	Word *p=(Word*)fetch(a,0);
	if((p->tag & 0x0f)!=TagWord)
		;//printf("\nerror: deref");
	return &(p->data);
}
uint16_t lookup(uint16_t env, uint16_t id) {
	return	env==nilAdrs? nilAdrs:
			caar(env)==id? cdar(env): lookup(cdr(env), id);
}
void initEnv() {
	uint16_t *p;
	int i;

	nilAdrs=newPair(0,0);
	p=&sym.nil;
	for(i=0; &p[i]<=&sym.last; i++)
		p[i]=name((char*)i);

	baseEnv=list(33,
				cons(sym.nil, nilAdrs),
				cons(sym.halt, sym.halt),
				cons(sym.add, sym.add),
				cons(sym.sub, sym.sub),
				cons(sym.mul, sym.mul),
				cons(sym.div, sym.div),

				cons(sym.equal, sym.equal),
				cons(sym.gt, sym.gt),
				cons(sym.ge, sym.ge),
				cons(sym.lt, sym.lt),
				cons(sym.le, sym.le),

				cons(sym.and_, sym.and_),
				cons(sym.or_, sym.or_),
				cons(sym.not_, sym.not_),


				cons(sym.cons, sym.cons),
				cons(sym.car, sym.car),
				cons(sym.cdr, sym.cdr),
				cons(sym.list, sym.list),
				cons(sym.quote, sym.quote),
				cons(sym.set, sym.set),
				cons(sym.setcar, sym.setcar),
				cons(sym.setcdr, sym.setcdr),

				cons(sym.io, sym.io),
				cons(sym.print, sym.print),
				cons(sym.send, sym.send),
				cons(sym.id, sym.id),
				cons(sym.msgq, sym.msgq),

				cons(sym.sleep, sym.sleep),
				cons(sym.eval, sym.eval),
				cons(sym.callcc, sym.callcc),
				cons(sym.seq, sym.seq),		//sym.par is caught directly from eval
				cons(sym.prim, sym.prim),
				cons(sym.extmsg, sym.extmsg)
				);// nilAdrs);
}
uint16_t haltCont(uint16_t id) {
	return list(4, sym.halt, nilAdrs, id, list(1, nilAdrs));	//tag, next continuation, id, message queue
}
static uint16_t bottomCont(uint16_t cont) {
	uint16_t next=cadr(cont);
	return next==nilAdrs? cont: bottomCont(next);
}
uint16_t getId(uint16_t cont) {
	return caddr(bottomCont(cont));
}
uint16_t getMsgQ(uint16_t cont) {
	return cadddr(bottomCont(cont));
}
uint16_t _reverse(uint16_t res, uint16_t l) {
	return l==nilAdrs? res: _reverse(cons(car(l), res), cdr(l));
}
uint16_t reverse(uint16_t l) {
	return _reverse(nilAdrs, l);
}
uint8_t isPair(uint16_t a) {
	return (*fetch(a,0) & 0x0f)==TagPair;
}
uint8_t isVariable(uint16_t a) {
	return (*fetch(a,0) & 0x0f)==TagSymbol;
}
uint8_t isValue(uint16_t a) {
	uint8_t tag=*fetch(a,0) & 0x0f;
	return tag==TagWord || tag==TagDWord || tag==TagFloat;
}
uint16_t _add(uint16_t args) {
	return args==nilAdrs? 0: *((uint16_t*)deref(car(args)))+_add(cdr(args));
}
uint16_t add(uint16_t args) {
	return newWord(_add(args));
}
uint16_t sub(uint16_t args) {
	return newWord(*((uint16_t*)deref(car(args)))-_add(cdr(args)));
}
uint16_t _mul(uint16_t args) {
	return args==nilAdrs? 1: *((uint16_t*)deref(car(args)))*_mul(cdr(args));
}
uint16_t mul(uint16_t args) {
	return newWord(_mul(args));
}
uint16_t divvy(uint16_t args) {
	return newWord(*((uint16_t*)deref(car(args)))/_mul(cdr(args)));
}
uint16_t io(uint16_t args) {
	uint16_t a=*((uint16_t*)deref(car(args)));
	uint16_t b=cdr(args)==nilAdrs? 0: *((uint16_t*)deref(car(cdr(args))));
	return newWord(IO(a, b));
}
uint16_t prim(uint16_t args) {
	uint16_t a=*((uint16_t*)deref(car(args)));
	uint16_t b=cdr(args);
	return PRIM(a, b);
}
uint16_t extmsg(uint16_t args) {
	printmsg(args);
	return car(args);
}
uint16_t pair(uint16_t args) {
	return cons(car(args), cadr(args));
}
uint16_t first(uint16_t args) {
	return caar(args);
}
uint16_t second(uint16_t args) {
	return cdar(args);
}
uint16_t setBang(uint16_t args) {
	return setdata(car(args), cadr(args));
}
uint16_t setFirst(uint16_t args) {
	return setcar(car(args), cadr(args));
}
uint16_t setSecond(uint16_t args) {
	return setcdr(car(args), cadr(args));
}
uint16_t eq(uint16_t args) {
	return car(args)==cadr(args)? sym.true: sym.false;
}
uint16_t equal(uint16_t args) {
	return *((uint16_t*)deref(car(args)))==*((uint16_t*)deref(cadr(args)))? sym.true: sym.false;
}
uint16_t gt(uint16_t args) {
	return *((uint16_t*)deref(car(args)))>*((uint16_t*)deref(cadr(args)))? sym.true: sym.false;
}
uint16_t ge(uint16_t args) {
	return *((uint16_t*)deref(car(args)))>=*((uint16_t*)deref(cadr(args)))? sym.true: sym.false;
}
uint16_t lt(uint16_t args) {
	return *((uint16_t*)deref(car(args)))<*((uint16_t*)deref(cadr(args)))? sym.true: sym.false;
}
uint16_t le(uint16_t args) {
	return *((uint16_t*)deref(car(args)))<=*((uint16_t*)deref(cadr(args)))? sym.true: sym.false;
}
uint16_t and_(uint16_t args) {
	return car(args)==sym.true && cadr(args)==sym.true? sym.true: sym.false;
}
uint16_t or_(uint16_t args) {
	return car(args)==sym.true || cadr(args)==sym.true? sym.true: sym.false;
}
uint16_t not_(uint16_t args) {
	return car(args)==sym.true? sym.false: sym.true;
}
uint16_t _print(uint16_t args) {
	if(args==nilAdrs)
		;//printf("\n");
	else {
		ledSet(*((uint16_t*)deref(car(args))));//printf("%d, ", deref(car(args)));
		_print(cdr(args));
	}
	return 0;
}
uint16_t print(uint16_t args) {
	_print(args);
	return car(args);
}
void apply(uint16_t cont, uint16_t v) {
	uint16_t tag;
	newGCRoot();
	tag=car(cont);
	if(tag==sym.halt) {					//(sym.halt)
		;
	}
	else if(tag==sym.cond_) {			//(sym.cond, nextCont, te, fe, env)
		uint16_t texp, fexp, env, c;
		c=cdr(cont); texp=cdr(c); fexp=cdr(texp); env=cdr(fexp);
		eval(v==sym.true? car(texp): car(fexp),  car(env), car(c));
	}
	else if(tag==sym.apply) {			//(sym.apply, nextCont, arg, res, env); arg: to be evaluated, res: already evaluated
		uint16_t arg, res, env, c, op;
		c=cdr(cont); arg=cdr(c); res=cdr(arg); env=cdr(res);
		c=car(c); arg=car(arg); res=cons(v, car(res)); env=car(env);
		if(arg!=nilAdrs) { 		//if there are more parameters to evaluate
			eval(car(arg), env, list(5, sym.apply, c, cdr(arg), res, env));
		}
		else {					//if all parameters are evaluated
			res=reverse(res);	//res after reverse: (op, arg1, arg2, ...)
			op=car(res);
			/////////////////////////////////////////////////////////////////
			//user defined func or continuation from callcc
			if(isPair(op)) {	
				tag=car(op);
				if(tag==sym.apply || tag==sym.cond_ || tag==sym.halt)	//continuation from callcc
					addThread(op, cadr(res));
				else if(tag==sym.closure) {								//user defined function
					uint16_t body=cdr(op), ids=cdr(body), envr=cdr(ids);
					eval(car(body), extendEnv(car(envr), car(ids), cdr(res)), c);
				}
#ifdef DEBUG
				else {
					fprintf(stderr,"error. apply: unknown pair tag %d\n", tag),
					exit(1);
				}
#endif
			}
			/////////////////////////////////////////////////////////////////
			//predefined functions
			else if(op==sym.add)	addThread(c, add(cdr(res)));
			else if(op==sym.sub)	addThread(c, sub(cdr(res)));
			else if(op==sym.mul)	addThread(c, mul(cdr(res)));
			else if(op==sym.div)	addThread(c, divvy(cdr(res)));

			else if(op==sym.equal)	addThread(c, equal(cdr(res)));
			else if(op==sym.gt)		addThread(c, gt(cdr(res)));
			else if(op==sym.ge)		addThread(c, ge(cdr(res)));
			else if(op==sym.lt)		addThread(c, lt(cdr(res)));
			else if(op==sym.le)		addThread(c, le(cdr(res)));

			else if(op==sym.and_)	addThread(c, and_(cdr(res)));
			else if(op==sym.or_)	addThread(c, or_(cdr(res)));
			else if(op==sym.not_)	addThread(c, not_(cdr(res)));

			else if(op==sym.cons)	addThread(c, pair(cdr(res)));
			else if(op==sym.car)	addThread(c, first(cdr(res)));
			else if(op==sym.cdr)	addThread(c, second(cdr(res)));
			else if(op==sym.list)	addThread(c, cdr(res));	//arg is already a list
			else if(op==sym.set)	addThread(c, setBang(cdr(res)));
			else if(op==sym.setcar)	addThread(c, setFirst(cdr(res)));
			else if(op==sym.setcdr)	addThread(c, setSecond(cdr(res)));

			else if(op==sym.io)		addThread(c, io(cdr(res)));
			else if(op==sym.print)	addThread(c, print(cdr(res)));
			else if(op==sym.send) {					//address, data should be added
				encode(cadr(res), *(uint16_t*)deref(getId(cont)));
				addThread(c, cadr(res));
			}
			else if(op==sym.id)		addThread(c, getId(cont));
			else if(op==sym.msgq)	addThread(c, getMsgQ(cont));

			else if(op==sym.sleep)	addThreadInState(c, nilAdrs, ThreadSleep);

			else if(op==sym.eval)	eval(cadr(res), baseEnv, c);
			//else if(op==sym.callcc) addThread(list(5, sym.apply, haltCont(getId(cont)), nilAdrs, cdr(res), env), c);
			else if(op==sym.callcc) addThread(list(5, sym.apply, bottomCont(cont), nilAdrs, cdr(res), env), c);
			else if(op==sym.seq)	addThread(c, v);	//continue with last expr
			else if(op==sym.prim)	addThread(c, prim(cdr(res)));
			else if(op==sym.extmsg)	addThread(c, extmsg(cdr(res)));
#ifdef DEBUG
			else {
				fprintf(stderr,"error. apply_2: %d\n", op);
				printlist(res);
				exit(1);
			}
#endif
		}
	}
}

void eval(uint16_t expr, uint16_t env, uint16_t cont) {
	if(isValue(expr)) {
		addThread(cont, expr);
	}
	else if(isVariable(expr)) {
		addThread(cont, lookup(env, expr));
	}
	else if(isPair(expr)) {
		uint16_t op=car(expr);
		if(op==sym.lambda) {
			uint16_t ids, body;
			ids=cdr(expr); body=cdr(ids);
			addThread(cont, list(4, sym.closure, car(body), car(ids), env));	//closure: body, ids, env
		}
		else if(op==sym.rec) {
			uint16_t ids, body, closure;
			ids=cdr(expr); body=cdr(ids);
			closure=list(4, sym.closure, car(body), cdr(car(ids)), env);	//closure: body, ids(wo self), env(w self)
			setcar(cdddr(closure), extendEnv(env, list(1, car(car(ids))), list(1, closure)));
			addThread(cont, closure);
		}
		else if(op==sym.cond) {
			uint16_t cexp, texp, fexp;
			cexp=cdr(expr); texp=cdr(cexp); fexp=cdr(texp);
			eval(car(cexp), env, list(5, sym.cond_, cont, car(texp), car(fexp), env));	//(tag, nextCont, true, false, env)
		}
		else if(op==sym.par) {	//launch threads to evaluate arguments
			uint16_t e, id, r=nilAdrs;
			for(e=cdr(expr); e!=nilAdrs; e=cdr(e)) {
				id=newWord(hwId()+thrId++);
				r=cons(id, r);
				eval(car(e), env, haltCont(id));
			}
			apply(cont, reverse(r));
		}
		else if(op==sym.halt || op==sym.cond_ || op==sym.apply){ 	//continuation as a one arg func.
			addThread(cont, expr);
		}
		else if(op==sym.quote) {
			addThread(cont, cadr(expr));
		}
		else if(op==sym.closure) {
			addThread(cont, expr);
		}
		else { //function application
			eval(car(expr), env, list(5, sym.apply, cont, cdr(expr), nilAdrs, env));	//(tag, cont, args, res, env)
		}
	}
	else {
		ledSet(7);
	}
}
