/**********************************************************
	extcomm.c
	version 1.0
	Sergei Shevlyagin (shevlyag@uiuc.edu)
**********************************************************/
#ifdef PC
#include <stdio.h>
#include <stdlib.h>

#include <time.h>
#include <signal.h>
#include "thread.h"
#endif
#include <string.h>
#include "data.h"
#include "extcomm.h"
#include "comm.h"
#include "memory.h"

#ifdef PC
timer_t timer, throttleTimer;
uint8_t timer_flag, throttleCount, seconds;
extern pthread_mutex_t rx_mt;
#endif

void(*sP)(uint8_t *, uint8_t);

void initSP(void(*newsP)(uint8_t *, uint8_t))
{
	sP = newsP;
}

#ifdef PC
void debugPrint(uint8_t *debugname, uint8_t *data, uint8_t n)
{
	uint8_t i; 
	if(Debug == 1)
	{
		fprintf(stderr, "|%s| %d ", debugname, n);
		for(i = 0; i<n;i++)
			fprintf(stderr, "%d ", data[i]);
		fprintf(stderr, "\n");
	}
}

void debugAck(uint8_t *text, acknowledgement ack)
{
	if(Debug == 1)
	{
		fprintf(stderr, "%s, Ack rxId:%d rxSeqNo:%d txId:%d txSeqNo:%d\n", text, ack.rxId, ack.rxSeqNo, ack.txId, ack.txSeqNo);
	}

}
#endif

void sendPacketAck(uint8_t *buf, uint8_t n, lastMessage lm[], uint8_t lmi)
{
	uint8_t giveup;
	giveup = 0;
	// setup flag
	lm[lmi].ok = 0;
	memcpy(&(lm[lmi].senderId), buf, 2);
	lm[lmi].seqNo = buf[2];
	
	setTimerFlag(0);
	setup_timer_resend(1);
	
	// try sending
#ifdef PC
	if(!message_troll())
	{
		sP(buf, n);
		debugPrint("sendPacket1", buf, n);
	}
#else
	sP(buf,n);
#endif

	// continue resending until acknowledged
	while(lm[lmi].ok == 0 && giveup != GiveUp)
	{
		if(getTimerFlag() == 1)
		{
			
			setTimerFlag(0);
			
			setup_timer_resend(1);
#ifdef PC
			if(!message_troll())
			{
				giveup++; 
				sP(buf, n);
				debugPrint("sendPacket2", buf, n);
			}
#else
			giveup++; 
			sP(buf,n);
#endif
		}
		
	}
	
	kill_timer_resend();
}
// PC timer functions

#ifdef PC
void setup_timer_resend(uint8_t delay)
{
	
	struct sigevent timer_event;
	struct itimerspec timer_time;
	struct sigaction act;
	
	if (delay>10)
		delay = 10; 

	// clear timer flag
	//timer_flag = flag;
	setTimerFlag(0);
	// assign our handler to SIGUSR1
	
	act.sa_flags = 0;
	act.sa_handler = timer_handler_resend;
	sigemptyset(&act.sa_mask);
	sigaction(SIGUSR1, &act, NULL);

	// our timer will signal SIGUSR1 instead of SIGALRM
	
	memset(&timer_event, 0, sizeof(timer_event));
	timer_event.sigev_notify = SIGEV_SIGNAL;
	timer_event.sigev_signo = SIGUSR1;
	
	timer_create(CLOCK_REALTIME, &timer_event, &timer);
	
	
	
	// set the timer
	timer_time.it_interval.tv_sec = delay;
	timer_time.it_interval.tv_nsec = 0;
	timer_time.it_value.tv_sec = delay;
	timer_time.it_value.tv_nsec = 0;
	timer_settime(timer, 0, &timer_time, NULL);
	//fprintf(stderr, "Armed\n");
	
}

void kill_timer_resend()
{
		struct itimerspec value;
	value.it_interval.tv_sec = 0;
	value.it_interval.tv_nsec = 0;
	value.it_value.tv_sec = 0;
	value.it_value.tv_nsec = 0;

	// start the timer
	timer_settime(timer, 0, &value, NULL);
	
	//fprintf(stderr, "DISarmed\n");
}

void timer_handler_resend()
{
	//fprintf(stderr, "TImer!\n");
	setTimerFlag(1);
	kill_timer_resend();
}

uint8_t getTimerFlag()
{
	return timer_flag;
}
void setTimerFlag(uint8_t val)
{
	timer_flag = val;
}
void setup_timer_throttle()
{
	
	struct sigevent timer_event;
	struct itimerspec timer_time;
	struct sigaction act;
	//fprintf(stderr, "setup timer \n");

	// assign our handler to SIGUSR1
	throttleCount = Throttle; 
	act.sa_flags = 0;
	act.sa_handler = timer_handler_throttle;
	sigemptyset(&act.sa_mask);
	sigaction(SIGUSR2, &act, NULL);

	// our timer will signal SIGUSR1 instead of SIGALRM
	
	memset(&timer_event, 0, sizeof(timer_event));
	timer_event.sigev_notify = SIGEV_SIGNAL;
	timer_event.sigev_signo = SIGUSR2;
	
	timer_create(CLOCK_REALTIME, &timer_event, &throttleTimer);
	
	
	
	// set the timer
	timer_time.it_interval.tv_sec = 1;
	timer_time.it_interval.tv_nsec = 0;
	timer_time.it_value.tv_sec = 1;
	timer_time.it_value.tv_nsec = 0;
	timer_settime(throttleTimer, 0, &timer_time, NULL);
	
}

void kill_timer_throttle()
{
	struct itimerspec value;
	value.it_interval.tv_sec = 0;
	value.it_interval.tv_nsec = 0;
	value.it_value.tv_sec = 0;
	value.it_value.tv_nsec = 0;
	//fprintf(stderr, "kill timer \n");
	// start the timer
	timer_settime(throttleTimer, 0, &value, NULL);
	//fprintf(stderr, "DISarmed\n");
}

void timer_handler_throttle()
{
	
	//fprintf(stderr, "shoot timer \n");
	seconds++; 
	throttleCount = Throttle; 
	
}


uint8_t message_troll()
{
	uint8_t n;
	n=random() %100;  // 0-99
	if (n>TrollLevel || (throttleCount == 0))
	{
		if(Debug == 1)
		{
			fprintf(stderr, "TROLL-ed\n");
		}
		return 1;
	}
	else
	{
		throttleCount--; 
		return 0;
	}
	
}
#endif
