/**********************************************************
	thread.c
	version 1.1
	YoungMin Kwon (ykwon4@cs.uiuc.edu)
**********************************************************/
#ifdef PC
#include <stdio.h>
#include <sched.h>
#endif
#include <setjmp.h>
#include "memory.h"
#include "eval.h"
#include "thread.h"
#include "hw.h"

Thread thr;

static uint8_t next(uint8_t p) {
	return (p+1)%MaxThread;
}
void initThread() {
	thr.h=thr.t=0;
	thr.next=next;
	thr.yielded=0;
}
void addThread(uint16_t cont, uint16_t value) {
	addThreadInState(cont, value, ThreadReady);
}
void addThreadInState(uint16_t cont, uint16_t value, uint8_t state) {
#ifdef DEBUG
	if(value==0)
		fprintf(stderr, "error. addThread: null point\n");
#endif
	if(next(thr.h)!=thr.t) {
		thr.cont[thr.h]=cont;
		thr.value[thr.h]=value;
		thr.state[thr.h]=state;
		thr.h=next(thr.h);
	}
#ifdef DEBUG
	else
		fprintf(stderr, "error. addThread: thread buffer full\n");
#endif
}
uint8_t *nullptr;
void runThread() {     
	uint8_t i=0, j=0;
	while(thr.h!=thr.t) {
		if(*nullptr!=0)	{								//memory error!
#ifdef DEBUG
			fprintf(stderr, "error. null point assignment\n");
#endif
			ledSet(7);
		}
#ifdef PC
#if defined(__CYGWIN__)
		sched_yield();
#else
		pthread_yield();
#endif
#endif

		if(!isRawBufEmpty()){ 							//comm
			saveStream();
		}
		if(thr.state[thr.t]==ThreadReady) {				//ready
			apply(thr.cont[thr.t], thr.value[thr.t]);
		}
		else {											//sleep
			addThreadInState(thr.cont[thr.t], thr.value[thr.t], thr.state[thr.t]);
		}
		thr.t=next(thr.t);
		if(i++>100) {
			i=0;
			yield();
			if(j++>100) {
				j=0;
				wakeupThreads();
			}
		}
	}
}
void wakeupThreads() {
	uint8_t i=0;
	for(i=thr.t; i!=thr.h; i=next(i)) {
		thr.state[i]=ThreadReady;
	}
}

void yield()
{
#ifdef STATT
	int res;
	printf("\r\t\t\tyield : %p", &res);
#endif
	if(setjmp(thr.toCont)==0) {
		longjmp(thr.toTos,1);
	}
}

void stackBottom()
{
	uint8_t stack[thr.stack_gap_size];	//stack space for tinyOs
#ifdef STATT
	printf("\r\t\t\t\t\t\tbottom : %p, %p", &stack[0], &stack[thr.stack_gap_size-1]);
#endif
	nullptr=&stack[0];
	*nullptr=0;
	thr.start(thr.start_arg);
	runThread();
	longjmp(thr.toTos, 2);			//return to current resume point
}

uint8_t resume()
{
	int res=setjmp(thr.toTos);		//return 0 if returns directly, non-zero otherwise
#ifdef STATT
	printf("\rresume: %p", &res);
#endif

	if(res!=0) {
		thr.yielded=1;
		return res!=2;				//return 0 if it is from stackBottom
	}
	else {
		if(thr.yielded) {
			thr.yielded=0;
			longjmp(thr.toCont,1);
		}
		else {
			stackBottom();
		}
		return 0;	
	}
}

void run(void (*start)(void*), void *arg, uint16_t stack_gap_size)
{
	thr.start=start;
	thr.start_arg=arg;
	thr.stack_gap_size=stack_gap_size;
}

